/*
	File:	driver.c
	Author: John van Groningen
*/

/* for parallel changes see _PARALLEL_ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <signal.h>

#include <sys/file.h>
#include <sys/stat.h>

extern char *getenv();

#define VERSION 840

/* file and directory names which may be patched using patchbin */

char clean_lib_directory[129] = ".";
char assembler_file_name[129] = "/bin/as";
char linker_file_name[129] = "/bin/ld";
char crt0_file_name [129] = "/lib/crt0.o";
char Mcrt1_file_name [129] = "/lib/Mcrt1.o";

/* default options of cocl */

#define DEFAULT_WARNING 				1
#define DEFAULT_INLINE					1
#define DEFAULT_VERBOSE 				0
#define DEFAULT_DEBUG					0
#define DEFAULT_TYPE_CHECK				1
#define DEFAULT_STRICTNESS_ANALYSIS 	1
#define DEFAULT_LIST_TYPES				0
#define DEFAULT_LIST_ALL_TYPES			0

#define DEBUG_MASK					1
#define INLINE_MASK 				2
#define PARALLEL_MASK				4
#define STACK_LAYOUT_MASK			8
#define STRICTNESS_ANALYSIS_MASK	16
#define TYPE_CHECK_MASK 			32
#define VERBOSE_MASK				64
#define WARNING_MASK				128
#define SYSTEM_MASK 				256

#define LIST_TYPES_MASK 			512
#define LIST_ALL_TYPES_MASK 		1024

#define DEFAULT_OPTIONS \
	(DEFAULT_WARNING ? WARNING_MASK : 0)\
	| (DEFAULT_INLINE ? INLINE_MASK : 0)\
	| (DEFAULT_VERBOSE ? VERBOSE_MASK : 0)\
	| (DEFAULT_DEBUG ? DEBUG_MASK : 0)\
	| (DEFAULT_TYPE_CHECK ? TYPE_CHECK_MASK : 0)\
	| (DEFAULT_STRICTNESS_ANALYSIS ? STRICTNESS_ANALYSIS_MASK : 0)\
	| (DEFAULT_LIST_TYPES ? LIST_TYPES_MASK : 0)\
	| (DEFAULT_LIST_ALL_TYPES ? LIST_ALL_TYPES_MASK : 0)

#define LEON

static int clean_options=DEFAULT_OPTIONS;	
static int clean_options_mask=0;
static int remove_symbol_table;
static int list_strict_export_types;

#ifdef sparc
static int stack_overflow_checks;
#endif

typedef struct project_node {
	struct project_node *	pro_next;
	char *					pro_fname;
	struct dep_list *		pro_depend;
	int 					pro_options;
	long					pro_abc_time;
	long					pro_dcl_time;
	unsigned int			pro_up_to_date:1,
							pro_valid_abc_time:1,
							pro_valid_dcl_time:1,
							pro_no_abc_file:1,
							pro_no_dcl_file:1,
							pro_valid_options:1
#ifdef LEON
						,	pro_ignore_o:1
#endif
							;
} *P_NODE;

typedef struct dep_list {
	struct dep_list *		dep_next;
	P_NODE					dep_node;
} *DEP_LIST;

P_NODE first_project_node,last_project_node;

struct library_list {
	struct library_list *	library_next;
	char					library_file_name[0];
};

#ifdef LEON
struct ignored_list {
	struct ignored_list * ignored_next;
	char				  ignored_file_name[0];
};

static struct ignored_list *first_ignored,*last_ignored;
#endif

static struct library_list *first_library,*last_library;

void warning (char *error_string)
{
	printf ("%s\n",error_string);
}

void warning_s (char *warning_string,char *s)
{
	printf (warning_string,s);
	putchar ('\n');
}

void warning_s_s (char *warning_string,char *s1,char *s2)
{
	printf (warning_string,s1,s2);
	putchar ('\n');
}

void error (char *error_string)
{
	printf ("%s\n",error_string);
	exit (1);
}

void error_s (char *error_string,char *s)
{
	printf (error_string,s);
	putchar ('\n');
	exit (1);
}

void *memory_allocate (int size)
{
	void *block;
	
	block=malloc (size);
	if (block==NULL)
		error ("Out of memory");
	return block;
}

P_NODE add_project_node (char *module_name)
{
	P_NODE new_project_node;
	char *pro_fname;
#ifdef LEON
	register struct ignored_list *ignored_o_modules;
#endif
	
	new_project_node=memory_allocate (sizeof (struct project_node));
	new_project_node->pro_next=NULL;
	
	pro_fname=memory_allocate (1+strlen (module_name));
	strcpy (pro_fname,module_name);

	new_project_node->pro_fname=pro_fname;
	new_project_node->pro_depend=NULL;
	new_project_node->pro_up_to_date=0;
	new_project_node->pro_valid_abc_time=0;
	new_project_node->pro_valid_dcl_time=0;
	new_project_node->pro_no_abc_file=0;
	new_project_node->pro_no_dcl_file=0;

#ifdef LEON
	new_project_node->pro_ignore_o=0;

	for (ignored_o_modules=first_ignored;
		ignored_o_modules!=NULL; ignored_o_modules=ignored_o_modules->ignored_next)
	{
		if (!strcmp (ignored_o_modules->ignored_file_name,module_name)){
			new_project_node->pro_ignore_o=1;
			break;
		}
	}
#endif

	new_project_node->pro_valid_options=0;
	
	if (first_project_node==NULL)
		first_project_node=new_project_node;
	else
		last_project_node->pro_next=new_project_node;
	last_project_node=new_project_node;
	
	return new_project_node;
}

P_NODE add_dependency (DEP_LIST *dependency_list_p,char *dependent_module)
{
	DEP_LIST new_dependency,dependency;
	P_NODE project_node;

	for (dependency=*dependency_list_p; dependency!=NULL;
		 dependency=dependency->dep_next)
	{
		if (!strcmp (dependent_module,dependency->dep_node->pro_fname))
			return NULL;
	}
	
	new_dependency=memory_allocate (sizeof (struct dep_list));
	
	for (project_node=first_project_node; project_node!=NULL;
		 project_node=project_node->pro_next)
	{
		if (!strcmp (dependent_module,project_node->pro_fname))
			break;
	}
	if (project_node==NULL)
		project_node=add_project_node (dependent_module);
	
	new_dependency->dep_node=project_node;
	
	new_dependency->dep_next=*dependency_list_p;
	*dependency_list_p=new_dependency;
	
	return project_node;
}

static char *clean_path_list,*clean_abc_path,*clean_o_path,*clean_lib_path;
static char clean_path_list_copy [513],clean_abc_path_copy[129];
static char clean_o_path_copy[129],clean_lib_path_copy[129];

void get_paths (void)
{
	clean_path_list=getenv ("CLEANPATH");		
	if (clean_path_list!=NULL)
		clean_path_list=strcpy (clean_path_list_copy,clean_path_list);
	
	clean_abc_path=getenv ("CLEANABCPATH");
	if (clean_abc_path!=NULL)
		clean_abc_path=strcpy (clean_abc_path_copy,clean_abc_path);
	
	clean_o_path=getenv ("CLEANOPATH");
	if (clean_o_path!=NULL)
		clean_o_path=strcpy (clean_o_path_copy,clean_o_path);
	
	clean_lib_path=getenv ("CLEANLIB");
	if (clean_lib_path!=NULL)
		clean_lib_path=strcpy (clean_lib_path_copy,clean_lib_path);
	else
		clean_lib_path=clean_lib_directory;
}

static int verbose=0;
static int silent=0;

int find_file
	(char *file_name,char *extension,char *complete_file_name,char *first_path)
{
	if (file_name[0]!='/'){
		if (first_path!=NULL){
			strcpy (complete_file_name,first_path);
			strcat (complete_file_name,"/");
			strcat (complete_file_name,file_name);
			strcat (complete_file_name,extension);
				
			if (access (complete_file_name,F_OK)==0)
				return 1;
		}
		
		if (clean_path_list!=NULL){
			char *s;
			int c;
			
			s=clean_path_list;
			c=*s++;
			while (c!='\0'){
				char *p;
				
				p=complete_file_name;
				while (c!=':' && c!='\0'){
					*p++=c;
					c=*s++;
				}
				*p++='/';
				strcpy (p,file_name);
				strcat (p,extension);
				
				if (access (complete_file_name,F_OK)==0)
					return 1;
				
				if (c==':')
					c=*s++;
			}
		}
	}
	
	strcpy (complete_file_name,clean_lib_path);
	strcat (complete_file_name,"/");
	strcat (complete_file_name,file_name);
	strcat (complete_file_name,extension);
	if (access (complete_file_name,F_OK)==0)
		return 1;
	
	strcpy (complete_file_name,file_name);
	strcat (complete_file_name,extension);
	return access (complete_file_name,F_OK)==0;
}

int get_abc_time (P_NODE project_node,long *time_p)
{
	struct stat stat_buffer;
	char file_name[128];
	long time;
	
	if (project_node->pro_valid_abc_time)
		time=project_node->pro_abc_time;
	else {
		if (project_node->pro_no_abc_file)
			return 0;
		
		if (!find_file (project_node->pro_fname,".abc",file_name,clean_abc_path))
			return 0;
	
		if (stat (file_name,&stat_buffer)<0){
			project_node->pro_no_abc_file=1;
			return 0;
		}
	
		time=stat_buffer.st_mtime;
		
		project_node->pro_abc_time=time;
		project_node->pro_valid_abc_time=1;
	}
	
	*time_p=time;
	return 1;
}

int get_dcl_time (P_NODE project_node,long *time_p)
{
	struct stat stat_buffer;
	char file_name[128];
	long time;
	
	if (project_node->pro_valid_dcl_time)
		time=project_node->pro_dcl_time;
	else {
		if (project_node->pro_no_dcl_file)
			return 0;
			
		if (!find_file (project_node->pro_fname,".dcl",file_name,NULL))
			return 0;
	
		if (stat (file_name,&stat_buffer)<0){
			project_node->pro_no_dcl_file=1;
			return 0;
		}
	
		time=stat_buffer.st_mtime;
		project_node->pro_dcl_time=time;
		project_node->pro_valid_dcl_time=1;
	}
	
	*time_p=time;
	return 1;
}

int get_time (P_NODE project_node,char *extension,long *time_p,char *first_path)
{
	struct stat stat_buffer;
	char file_name[128];
	
	if (!find_file (project_node->pro_fname,extension,file_name,first_path))
		return 0;
	
	if (stat (file_name,&stat_buffer)<0)
		return 0;
	
	*time_p=stat_buffer.st_mtime;
	return 1;
}

FILE *abc_file;
int last_char;

int open_abc_file (P_NODE project_node)
{
	char file_name [128];
	FILE *f;
	
	if (!find_file (project_node->pro_fname,".abc",file_name,clean_abc_path))
		return 0;
	
	f=fopen (file_name,"r");
	if (f==NULL)
		return 0;
	
	last_char=getc (f);
	abc_file=f;
	return 1;
}

void skip_spaces (void)
{
	while (last_char==' ' || last_char=='\t')
		last_char=getc (abc_file);
}

void skip_to_next_line (void)
{
	while (last_char!='\n' && last_char!=EOF)
		last_char=getc (abc_file);
	if (last_char=='\n')
		last_char=getc (abc_file);
}

int last_char_was (int c)
{
	if (last_char==c){
		last_char=getc (abc_file);
		return 1;
	} else
		return 0;
}

int last_char_was_digit (void)
{
	if (last_char<='9' && last_char>='0'){
		last_char=getc (abc_file);
		return 1;
	} else
		return 0;
}

int get_version_and_options_of_abc_file (P_NODE project_node,int *version_p)
{
	int version,options;
	
	while (last_char!=EOF){
		skip_spaces();
		if (last_char!='.')
			return 0;
		last_char=getc (abc_file);
		if (last_char_was ('c')){
			if (last_char_was ('o') && last_char_was ('m') && last_char_was ('p')){
				int digit_1,digit_2,digit_3;
				
				skip_spaces();
				
				if ((digit_1=last_char,last_char_was_digit())
					&& (digit_2=last_char,last_char_was_digit())
					&& (digit_3=last_char,last_char_was_digit()))
				{
					int digit_n;
					
					version=(digit_1-'0')*100 + (digit_2-'0')*10 + (digit_3-'0');
					
					skip_spaces();
					
					options=0;
					if (last_char!='\n')
						for (digit_n=0; digit_n<9; ++digit_n){
							if (last_char=='0')
								;
							else if (last_char=='1')
								options |= (1<<digit_n);
							else
								return 0;
							last_char=getc (abc_file);
						}
					
					skip_to_next_line();
					
					project_node->pro_options=options;
					project_node->pro_valid_options=1;
					
					*version_p=version;
					return 1;
				}
			}
		} else if (last_char_was ('e') && last_char_was ('n') && last_char_was ('d')
			&& last_char_was ('i') && last_char_was ('n') && last_char_was ('f')
			&& last_char_was ('o'))
		{
			return 0;
		}
		skip_to_next_line();
	}
	return 0;
}

int get_dependency_of_abc_file (char *file_name,int max_length)
{
	while (last_char!=EOF){
		skip_spaces();
		if (last_char!='.')
			return 0;
		last_char=getc (abc_file);
		if (last_char_was ('d')){
			if (last_char_was ('e') && last_char_was ('p') && last_char_was ('e')
				&& last_char_was ('n') && last_char_was ('d'))
			{
				int n;
				
				skip_spaces();
				
				if (last_char=='\"'){
					last_char=getc (abc_file);
					n=0;
					while (last_char!='\"' && last_char!='\n' && last_char!=EOF
						   && n<max_length)
					{
						*file_name++=last_char;
						last_char=getc (abc_file);
						++n;
					}
					if (n<max_length && last_char=='\"'){
						*file_name='\0';
						skip_to_next_line();
						return 1;
					}
				}
			}
			return 0;
		} else if (last_char_was ('e') && last_char_was ('n') && last_char_was ('d')
			&& last_char_was ('i') && last_char_was ('n') && last_char_was ('f')
			&& last_char_was ('o'))
		{
			return 0;
		}
		skip_to_next_line();
	}
	return 0;
}

void close_abc_file (void)
{
	fclose (abc_file);
}

static int single_module=0;
static int syntax_check=0;
static int only_abc_files=0;

/* _PARALLEL_ */

static int only_s_files=1;
static int only_o_files=0;

int project_node_is_abc_up_to_date (P_NODE project_node)
{
	long abc_time,dcl_time,icl_time;
	int version,abc_options;
	DEP_LIST dependency_list;
	char dependent_module [128];
	
	if (!get_abc_time (project_node,&abc_time)){
		if (verbose)
			warning_s ("(%s.abc doesn't exist)",project_node->pro_fname);
		return 0;
	}
	
	if (!open_abc_file (project_node))
		return 0;
	
	if (!get_version_and_options_of_abc_file (project_node,&version)){
		close_abc_file();
		return 0;
	}
	abc_options=project_node->pro_options;
	
	if (version!=VERSION){
		close_abc_file();
		warning_s ("Warning: %s.abc is generated with another compiler version",
				   project_node->pro_fname);
		return 0;
	}
	
	if (!(abc_options & SYSTEM_MASK)){
		if (!(abc_options & STACK_LAYOUT_MASK)){
			if (verbose)
				warning_s ("(%s.abc does not contain stack information)",
						   project_node->pro_fname);
			close_abc_file();
			return 0;
		}
	
		if (get_time (project_node,".icl",&icl_time,NULL) && abc_time<=icl_time){
			if (verbose)
				warning_s ("(%s.abc is older than corresponding icl file)",
						   project_node->pro_fname);
			close_abc_file();
			return 0;
		}
		
		if (get_dcl_time (project_node,&dcl_time) && abc_time<=dcl_time){
			if (verbose)
				warning_s ("(%s.abc is older than corresponding dcl file)",
						   project_node->pro_fname);
			return 0;
		}
	}
	
	if (project_node==first_project_node &&
		((DEBUG_MASK | INLINE_MASK | STRICTNESS_ANALYSIS_MASK | TYPE_CHECK_MASK |
		  WARNING_MASK) & clean_options_mask & (clean_options ^ abc_options))!=0)
	{
		if (verbose)
			warning_s ("(%s.icl is compiled with different options)",
					   project_node->pro_fname);
		close_abc_file();
		return 0;
	}	

	dependency_list=NULL;
	while (get_dependency_of_abc_file (dependent_module,128)){
		long dcl_time;
		P_NODE dependency_node;
		
		dependency_node=add_dependency (&dependency_list,dependent_module);
		
		if (dependency_node==NULL)
			continue;
		
		if (!(abc_options & SYSTEM_MASK)){
			if (get_dcl_time (dependency_node,&dcl_time) && abc_time<=dcl_time){
				if (verbose)
					warning_s_s ("(%s.icl conflicts with %s.dcl)",project_node->pro_fname,
								 dependent_module);
				close_abc_file();
				return 0;
			}
			
			/*
			if (is_system_file (dependency_node)){
				long dcl_abc_time;
				
				if (get_abc_time (dependency_node,&dcl_abc_time) && abc_time<=dcl_abc_time){
					if (verbose)
						warning_s_s ("(%s.icl conflicts with %s.abc)",
									 project_node->pro_fname,dependency_node->pro_fname);
					close_abc_file();
					return 0;
				}
			}
			*/
		}
	}

	close_abc_file();
	
	project_node->pro_depend=dependency_list;
	
	if (verbose)
		warning_s ("(%s.abc is up to date)",project_node->pro_fname);
	return 1;
}

int is_system_file (P_NODE project_node)
{
	if (project_node->pro_no_abc_file)
		return 0;
	
	if (!project_node->pro_valid_options){
		int version;
		FILE *old_file;
		
		old_file=abc_file;
		
		if (!open_abc_file (project_node)){
			abc_file=old_file;
			return 0;
		}
	
		if (!get_version_and_options_of_abc_file (project_node,&version)){
			close_abc_file();
			abc_file=old_file;
			return 0;
		}
		close_abc_file();
		abc_file=old_file;
	}
	
	return (project_node->pro_options & SYSTEM_MASK)!=0;
}

int compile_project_node (P_NODE project_node)
{
	int pid,r,status;
	char file_name[128];
	char *abc_file_name,abc_file_name_s[128];
	char cocl_file_name[128];
	
	if (!silent)
		printf ("Compiling %s\n",project_node->pro_fname);
	
	project_node->pro_no_abc_file=0;
	project_node->pro_valid_abc_time=0;
	project_node->pro_valid_options=0;
	
	if (!find_file (project_node->pro_fname,".icl",file_name,NULL))
		error_s ("Can't find %s.icl",project_node->pro_fname);
	
	if (clean_abc_path==NULL)
		abc_file_name=file_name;
	else {
		abc_file_name=abc_file_name_s;
		strcpy (abc_file_name,clean_abc_path);
		strcat (abc_file_name,"/");
		strcat (abc_file_name,project_node->pro_fname);
		strcat (abc_file_name,".abc");
	}
	
	pid=fork();
	if (pid<0)
		error ("Fork failed");
	
	if (!pid){
		char *argv[16],**arg;
		int options;
		
		arg=argv;
		*arg++="cocl";
		*arg++="-sl";
		
		/* _PARALLEL_ */
		
		*arg++="-p";		/* generate parallel code */
		
		if (syntax_check)
			*arg++="-c";
		
		if (((clean_options & VERBOSE_MASK)!=0)!=DEFAULT_VERBOSE)
			*arg++="-v";
		if (((clean_options & LIST_TYPES_MASK)!=0)!=DEFAULT_LIST_TYPES)
			*arg++="-lt";
		if (((clean_options & LIST_ALL_TYPES_MASK)!=0)!=DEFAULT_LIST_ALL_TYPES)
			*arg++="-lat";
		
		if (project_node->pro_valid_options)
			options=project_node->pro_options;
		else
			options=DEFAULT_OPTIONS;
		
		if (project_node==first_project_node)
			options=(options & ~clean_options_mask) 
					| (clean_options & clean_options_mask);
		
		if (((options & WARNING_MASK)!=0)!=DEFAULT_WARNING)
			*arg++="-w";
		if (((options & INLINE_MASK)!=0)!=DEFAULT_INLINE)
			*arg++="-i";
		if (((options & DEBUG_MASK)!=0)!=DEFAULT_DEBUG)
			*arg++="-d";
		if (((options & TYPE_CHECK_MASK)!=0)!=DEFAULT_TYPE_CHECK)
			*arg++="-tc";
		if (((options & STRICTNESS_ANALYSIS_MASK)!=0)!=DEFAULT_STRICTNESS_ANALYSIS)
			*arg++="-sa";
	
		if (list_strict_export_types)
			*arg++="-lst";
		
		*arg++="-o";
		*arg++=abc_file_name;
		*arg++=file_name;
		*arg=0;
		
		strcpy (cocl_file_name,clean_lib_path);
		strcat (cocl_file_name,"/cocl");
		execv (cocl_file_name,argv);
		
		error ("Can't execute the clean compiler");
	}
	
	r=wait (&status);
	
	return r>=0 && status==0;
}

/* _PARALLEL_ */

int generate_code (P_NODE project_node,char *file_name)
{
	int pid,r,status;
	char cg_file_name[128];
	
	if (!silent)
		printf ("Generating transputer code for %s\n",project_node->pro_fname);
	
	pid=fork();
	if (pid<0)
		error ("Fork failed");
	
	if (!pid){
		char *argv[16],**arg;
		
		arg=argv;
		*arg++="tg";

		*arg++=file_name;
		*arg=0;
		
		strcpy (cg_file_name,clean_lib_path);
		strcat (cg_file_name,"/tg");
		execv (cg_file_name,argv);
		
		error ("Can't execute the transputer code generator");
	}
	
	r=wait (&status);
	return r>=0 && status==0;
}

int assemble (P_NODE project_node,char *file_name)
{
	int pid,r,status;
	char s_file_name[128],o_file_name[128];
	
	if (!silent)
		printf ("Assembling %s\n",project_node->pro_fname);
	
	error("transputer assembler not yet available");
	return 0;

#if 0

	strcpy (s_file_name,file_name);
	strcat (s_file_name,".s");
	
	if (clean_o_path==NULL)
		strcpy (o_file_name,file_name);
	else {
		strcpy (o_file_name,clean_o_path);
		strcat (o_file_name,"/");
		strcat (o_file_name,project_node->pro_fname);
	}
	strcat (o_file_name,".o");
	
	pid=fork();
	if (pid<0)
		error ("Fork failed");
	
	if (!pid){
		char *argv[16],**arg;
		
		arg=argv;
		*arg++="as";
		*arg++=s_file_name;
		*arg++="-o";
		*arg++=o_file_name;
		*arg=0;
		
		execv (assembler_file_name,argv);
		
		error ("Can't execute the assembler");
	}
	
	r=wait (&status);
	return r>=0 && status==0;
#endif
}

void get_dependencies (P_NODE project_node)
{
	DEP_LIST dependency_list;
	char dependent_module [128];
	int version;
	
	if (!open_abc_file (project_node))
		return;
	
	if (!get_version_and_options_of_abc_file (project_node,&version)){
		close_abc_file();
		return;
	}
	
	dependency_list=NULL;
	while (get_dependency_of_abc_file (dependent_module,128))
		add_dependency (&dependency_list,dependent_module);
	
	close_abc_file();
	
	project_node->pro_depend=dependency_list;
}

int make_project_to_abc_files (P_NODE project_node)
{
	DEP_LIST dependency;
	
	if (project_node->pro_up_to_date)
		return 1;
	
	project_node->pro_up_to_date=1;
	
	if (!project_node_is_abc_up_to_date (project_node)){
		if (is_system_file (project_node))
			return 0;
		if (!compile_project_node (project_node))
			return 0;
		if (!single_module)
			get_dependencies (project_node);
	}
	
	if (!single_module)
		for (dependency=project_node->pro_depend; dependency!=NULL;
			 dependency=dependency->dep_next)
		{
			if (!make_project_to_abc_files (dependency->dep_node))
				return 0;
		}
	
	return 1;
}

int project_node_is_o_up_to_date (P_NODE project_node)
{
	long abc_time,o_time;
	
	if (!get_time (project_node,".o",&o_time,
				   clean_o_path!=NULL ? clean_o_path : clean_abc_path))
	{
		if (verbose)
			warning_s ("(%s.o doesn't exist)",project_node->pro_fname);
		return 0;
	}
	
	if (get_abc_time (project_node,&abc_time) && o_time<=abc_time){
		if (verbose)
			warning_s ("(%s.o is older than corresponding abc file)",
					   project_node->pro_fname);
		return 0;
	}
	
	if (verbose)
		warning_s ("(%s.o is up to date)",project_node->pro_fname);
	return 1;
}

int make_project_to_o_files (void)
{
	P_NODE project_node;
	
	for (project_node=first_project_node; project_node!=NULL;
		 project_node=project_node->pro_next)
	{
		if (project_node->pro_up_to_date
#ifdef LEON
			&& !project_node->pro_ignore_o
#endif
			&& !project_node_is_o_up_to_date (project_node))
		{
			char file_name[128];
			
			if (!find_file (project_node->pro_fname,".abc",file_name,clean_abc_path))
				error_s ("Can't find %s.abc",project_node->pro_fname);
			file_name[strlen (file_name)-4]='\0';
			
			if (!generate_code (project_node,file_name))
				return 0;
			
			if (!assemble (project_node,file_name))
				return 0;
			
			strcat (file_name,".s");
			unlink (file_name);
		}
	}
	return 1;
}

int project_node_is_s_up_to_date (P_NODE project_node)
{
	long abc_time,s_time;
	
	if (!get_time (project_node,".s",&s_time,clean_abc_path)){
		if (verbose)
			warning_s ("(%s.s doesn't exist)",project_node->pro_fname);
		return 0;
	}
	
	if (get_abc_time (project_node,&abc_time) && s_time<=abc_time){
		if (verbose)
			warning_s ("(%s.s is older than corresponding abc file)",
					   project_node->pro_fname);
		return 0;
	}
	
	if (verbose)
		warning_s ("(%s.s is up to date)",project_node->pro_fname);
	return 1;
}

int make_project_to_s_files (void)
{
	P_NODE project_node;
	int abc_options;
	
	for (project_node=first_project_node; project_node!=NULL;
		 project_node=project_node->pro_next)
	{
		abc_options=project_node->pro_options;
		
		if (project_node->pro_up_to_date
			&& !project_node_is_s_up_to_date (project_node)
			&& !(abc_options & SYSTEM_MASK))
		{
			char file_name[128];
			
			if (!find_file (project_node->pro_fname,".abc",file_name,clean_abc_path))
				error_s ("Can't find %s.abc",project_node->pro_fname);
			file_name[strlen (file_name)-4]='\0';
			
			if (!generate_code (project_node,file_name))
				return 0;
		}
	}
	return 1;
}

long data [] =
#ifndef sparc
{
	0x00020107, 0x00000000, 0x00000010, 0x00000000,
	0x0000003c, 0x00000000, 0x00000000, 0x00000000,
	0x00200000, 0x00080000, 0x00080000, 0x00000008,
	0x00000004, 0x07000000, 0x0000000c, 0x0000000b,
	0x07000000, 0x00000010, 0x00000014, 0x07000000,
	0x00000000, 0x0000001f, 0x07000000, 0x00000008,
	0x0000002d, 0x07000000, 0x00000004, 0x0000003c,
	0x5f666c61, 0x67730064, 0x6174615f, 0x656e6400,
	0x5f686561, 0x705f7369, 0x7a65005f, 0x635f7374,
	0x61636b5f, 0x73697a65, 0x005f6162, 0x5f737461,
	0x636b5f73, 0x697a6500
};
#else
{
	0x01030107,0x00000000,0x00000010,0x00000000,
	0x0000003c,0x00000000,0x00000000,0x00000000,
	0x00200000,0x00080000,0x00080000,0x00000008,
	0x00000004,0x07000000,0x00000008,0x00000012,
	0x07000000,0x0000000c,0x00000019,0x07000000,
	0x00000004,0x00000028,0x07000000,0x00000000,
	0x00000033,0x07000000,0x00000010,0x0000003c,
	0x5f635f73,0x7461636b,0x5f73697a,0x65005f66,
	0x6c616773,0x005f6162,0x5f737461,0x636b5f73,
	0x697a6500,0x5f686561,0x705f7369,0x7a650064,
	0x6174615f,0x656e6400
};
#endif

long ab_stack_size,heap_size,flags;

int create_options_file (char **options_file_name_p)
{
	char *options_file_name;
	FILE *f;
	
	options_file_name=tempnam (NULL,"cgopt");
	if (options_file_name==NULL)
		return 0;
	
	f=fopen (options_file_name,"wb");
	if (f==NULL)
		error_s ("Can't create options file %s",options_file_name);
	
	data[8]=heap_size;
	data[9]=ab_stack_size;
	data[10]=512<<10;
	data[11]=flags;

	if (fwrite (data,1,sizeof (data),f)!=sizeof (data)){
		fclose (f);
		unlink (options_file_name);
		error ("Error writing options file");
	}
	
	if (fclose (f)!=0){
		unlink (options_file_name);
		error ("Error writing options file");
	}
	
	*options_file_name_p=options_file_name;
	return 1;
}

int link_project
	(P_NODE first_project_node,char *options_file_name,char *application_file_name)
{
	P_NODE project_node;
	int pid,r,status;
	char o_file_name[128],start_up_file_name[128],system_file_name[128];
	char *argv[128],**arg;
	struct library_list *library;
#ifdef sparc
	char reals_file_name[128];
#endif
	
	if (!silent)
		printf ("Linking %s\n",first_project_node->pro_fname);
	
	arg=argv;
	*arg++="ld";
#ifndef sparc
	*arg++="-N";
#endif
	*arg++="-e";
	*arg++="start";
	*arg++="-dc";
	*arg++="-dp";
	
	if (remove_symbol_table)
		*arg++="-s";
	
	if (application_file_name!=NULL){
		*arg++="-o";
		*arg++=application_file_name;
	}
	
	*arg++=crt0_file_name;
#ifndef sparc
	*arg++=Mcrt1_file_name;
#endif
	
	strcpy (start_up_file_name,clean_lib_path);
	strcat (start_up_file_name,"/_startup.o");
	*arg++=start_up_file_name;
	
	strcpy (system_file_name,clean_lib_path);
	strcat (system_file_name,"/_system.o");
	*arg++=system_file_name;
	
	for (project_node=first_project_node; project_node!=NULL;
		 project_node=project_node->pro_next)
	{
		if (project_node->pro_up_to_date
#ifdef LEON
		   && !(project_node->pro_ignore_o)
#endif
		   ){
			char *file_name;
			
			if (!find_file (project_node->pro_fname,".o",o_file_name,
							clean_o_path!=NULL ? clean_o_path : clean_abc_path))
			{
				unlink (options_file_name);
				error_s ("Can't find %s.o",project_node->pro_fname);
			}
				
			file_name=malloc (strlen (o_file_name)+1);
			if (file_name==NULL){
				unlink (options_file_name);
				error ("Out of memory");
			}
		
			strcpy (file_name,o_file_name);
			*arg++=file_name;
		}
	}

	for (library=first_library; library!=NULL; library=library->library_next)
		*arg++=library->library_file_name;

	*arg++=options_file_name;

#ifdef sparc
	strcpy (reals_file_name,clean_lib_path);
	strcat (reals_file_name,"/_reals.o");
	*arg++=reals_file_name;

	*arg++="-lm";
#endif

	*arg++="-lc";
	*arg=NULL;

	pid=fork();
	if (pid<0){
		unlink (options_file_name);
		error ("Fork failed");
	}
	
	if (!pid){
		execv (linker_file_name,argv);
		
		error ("Can't execute the linker");
	}
	
	r=wait (&status);
	return r>=0 && status==0;
}

void argument_error (void)
{
	printf ("Usage: clm [options] module_name [-o application_name]\n");
	printf ("project options: -mv -ms -O -S -ABC -PO -PS -PABC -c -lt -lat -ns -lst\n");
	printf ("application options: -h size -s size -b -sc -t -nt -gc -ngc -st -nst -nr\n");
	printf ("main module options: -w -i -d -tc -sa -v -nw -ni -nd -ntc -nsa -nv\n");
	exit (1);
}

long parse_size (register char *s)
{
	register int c;
	register long n;
	
	c=*s++;
	if (c<'0' || c>'9')
		error ("Digit expected in argument\n");
	
	n=c-'0';
	
	while (c=*s++,c>='0' && c<='9')
		n=n*10+(c-'0');
	
	if (c=='k' || c=='K'){
		c=*s++;
		n<<=10;
	} else if (c=='m' || c=='M'){
		c=*s++;
		n<<=20;
	}
	
	if (c!='\0')
		error ("Error in size");
	
	return n;
}

static void add_library (char *library_file_name)
{
	int file_name_length;
	register struct library_list *new_library_file;
	
	file_name_length=strlen (library_file_name);
	new_library_file=memory_allocate
		(sizeof (struct library_list)+file_name_length+1);
		
	strcpy (new_library_file->library_file_name,library_file_name);
	new_library_file->library_next=NULL;
	
	if (first_library==NULL)
		first_library=new_library_file;
	else
		last_library->library_next=new_library_file;
	last_library=new_library_file;
}

#ifdef LEON
static void add_ignored (char *ignored_name)
{
	int file_name_length;
	register struct ignored_list *new_ignored_file;

	file_name_length=strlen(ignored_name);
	new_ignored_file=memory_allocate (sizeof(struct ignored_list)+file_name_length+1);

	strcpy (new_ignored_file->ignored_file_name,ignored_name);
	new_ignored_file->ignored_next=NULL;

	if (first_ignored==NULL)
		first_ignored=new_ignored_file;
	else
		last_ignored->ignored_next=new_ignored_file;
	last_ignored=new_ignored_file;
}

static void read_shared_library (char *shared_script_name)
{
	char script_file_name[128];
	char file_name[128];
	FILE *fp;

	if (!find_file (shared_script_name,".lo",script_file_name,clean_lib_path))
		error_s ("Can't find %s.lo",shared_script_name);

	if (!(fp=fopen (script_file_name,"r")))
		error_s ("Can't open %s.lo",shared_script_name);

	while (fscanf (fp,"%s\n",file_name)!=EOF && file_name[0]!='=')
		add_library (file_name);

	while (fscanf (fp,"%s\n",file_name)!=EOF)
		add_ignored (file_name);
}
#endif

#define option_on(mask)  clean_options|=(mask);  clean_options_mask|=(mask);
#define option_off(mask) clean_options&=~(mask); clean_options_mask|=(mask);

int main (int argc,char **argv)
{
	int arg_n;
	char *application_file_name;
	
	if (argc<2 || argv[argc-1][0]=='-')
		argument_error();
	
	heap_size=2048<<10;
	ab_stack_size=512<<10;
	flags=8;
	
	remove_symbol_table=1;
#ifdef sparc
	stack_overflow_checks=0;
#endif

	first_library=NULL;
	last_library=NULL;

#ifdef LEON
	first_ignored=NULL;
	last_ignored=NULL;
	get_paths();
#endif
	
	for (arg_n=1; arg_n<argc && argv[arg_n][0]=='-'; ++arg_n){
		char *argument,*s;
		
		argument=argv[arg_n];		
		s=argument+1;
		
		/* driver options */
		
		if (!strcmp (s,"mv"))
			verbose=1;
		else if (!strcmp (s,"ms"))
			silent=1;
		else if (!strcmp (s,"c")){
			single_module=1;
			syntax_check=1;
		} else if (!strcmp (s,"ABC")){
			single_module=1;
			only_abc_files=1;
		} else if (!strcmp (s,"S")){
			single_module=1;
			only_s_files=1;
		} else if (!strcmp (s,"O")){
			single_module=1;
			only_o_files=1;
		} else if (!strcmp (s,"PABC"))
			only_abc_files=1;
		else if (!strcmp (s,"PS"))
			only_s_files=1;
		else if (!strcmp (s,"PO"))
			only_o_files=1;
		else if (!strcmp (s,"ns"))
			remove_symbol_table=0;

		/* application options */
		
		else if (!strcmp (s,"h")){
			++arg_n;
			if (arg_n>=argc)
				exit ("Heap size missing\n");
			heap_size=parse_size (argv[arg_n]);
		} else if (!strcmp (s,"s")){
			++arg_n;
			if (arg_n>=argc)
				error ("Stack size missing\n");
			ab_stack_size=parse_size (argv[arg_n]);
		} else if (!strcmp (s,"b"))
			flags |= 1;
		else if (!strcmp (s,"sc"))
			flags &= ~1;
		else if (!strcmp (s,"t"))
			flags |= 8;
		else if (!strcmp (s,"nt"))
			flags &= ~8;
		else if (!strcmp (s,"gc"))
			flags |= 2;
		else if (!strcmp (s,"ngc"))
			flags &= ~2;
		else if (!strcmp (s,"st"))
			flags |= 4;
		else if (!strcmp (s,"nst"))
			flags &= ~4;
		else if (!strcmp (s,"nr"))
			flags |= 16;

		/* clean options */
		
		else if (!strcmp (s,"v")){
			option_on (VERBOSE_MASK);
		} else if (!strcmp (s,"nv")){
			option_off (VERBOSE_MASK);
		} else if (!strcmp (s,"w")){
			option_on (WARNING_MASK);
		} else if (!strcmp (s,"nw")){
			option_off (WARNING_MASK);
		} else if (!strcmp (s,"i")){
			option_on (INLINE_MASK);
		} else if (!strcmp (s,"ni")){
			option_off (INLINE_MASK);
		} else if (!strcmp (s,"d")){
			option_on (DEBUG_MASK);
		} else if (!strcmp (s,"nd")){
			option_off (DEBUG_MASK);
		} else if (!strcmp (s,"tc")){
			option_on (TYPE_CHECK_MASK);
		} else if (!strcmp (s,"ntc")){
			option_off (TYPE_CHECK_MASK);
		} else if (!strcmp (s,"sa")){
			option_on (STRICTNESS_ANALYSIS_MASK);
		} else if (!strcmp (s,"nsa")){
			option_off (STRICTNESS_ANALYSIS_MASK);
		} else if (!strcmp (s,"lt")){
			option_on (LIST_TYPES_MASK);
		} else if (!strcmp (s,"nlt")){
			option_off (LIST_TYPES_MASK);
		} else if (!strcmp (s,"lat")){
			option_on (LIST_ALL_TYPES_MASK);
		} else if (!strcmp (s,"nlat")){
			option_off (LIST_ALL_TYPES_MASK);
		} else if (!strcmp (s,"lst")){
			list_strict_export_types=1;
		
		/* code generator options */
/*		} else if (!strcmp (s,"oab")){
			stack_overflow_checks=1;
		} else if (!strcmp (s,"oc")){
			stack_overflow_checks=1;
*/
		} else if (!strcmp (s,"l")){
			++arg_n;
			if (arg_n>=argc)
				error ("Library name missing\n");
			add_library (argv[arg_n]);
#ifdef LEON
		} else if (!strcmp (s,"sl")){
			++arg_n;
			if (arg_n>=argc)
			   error ("Shared library name missing\n");
			read_shared_library (argv[arg_n]);
#endif
		} else
			error_s ("Unknown option: %s",argument);
	}
	
	if (arg_n==argc-3){
		if (strcmp (argv[arg_n+1],"-o")!=0)
			argument_error();
		application_file_name=argv[arg_n+2];
	} else {
		application_file_name=NULL;
		if (arg_n!=argc-1)
			argument_error();
	}
	
#ifndef LEON
	get_paths();
#endif

	add_project_node (argv[arg_n]);
	if (make_project_to_abc_files (first_project_node)){
		if (only_abc_files || syntax_check)
			return 0;
		
		if (only_s_files)
			return make_project_to_s_files();
		
		if (make_project_to_o_files()){
			char *options_file_name;
			
			if (only_o_files)
				return 0;
			
			if (create_options_file (&options_file_name)){
				if (!link_project
						(first_project_node,options_file_name,application_file_name))
				{
					unlink (options_file_name);
					return 1;
				}
				unlink (options_file_name);
				return 0;
			}
		}
	}
	
	return 1;
}
